SEARCHING_KEY_SIZE_MORE_THAN_DEFAULT_SEARCHING_KEY_ERROR = (
    'Ключ для поиска не может быть больше, чем ключ, на основе которого была '
    'построена хеш-таблица!'
)

SEARCHING_KEY_SIZE_LESS_THAN_DEFAULT_SEARCHING_KEY_ERROR = (
    'Ключ для поиска не может быть меньше, чем ключ, на основе которого была '
    'построена хеш-таблица в строгом режиме поиска!'
)

DATE_FROM_MORE_OR_EQUAL_DATE_TO_ERROR = (
    'Начальная дата должна быть строго меньше конечной даты'
)

