# Наименование модуля с функцией
FUNCTIONS_MODULE = 'functions'

# Путь модуля с базовыми функциями
FUNCTION_TOOLS_FUNCTIONS_MODULE_PATH = f'function_tools.{FUNCTIONS_MODULE}'

# Пустое значение полного наименования функции
EMPTY_VERBOSE_NAME = 'Empty verbose name'
