# Generated by Django 3.2.6 on 2022-02-16 13:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('function_tools', '0003_implementationstrategy'),
    ]

    operations = [
        migrations.AddField(
            model_name='implementationstrategy',
            name='order_number',
            field=models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер'),
        ),
    ]
