# Generated by Django 2.2.4 on 2021-09-05 23:05

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RegisteredFunction',
            fields=[
                ('function_path', models.CharField(db_index=True, max_length=1024, primary_key=True, serialize=False, unique=True, verbose_name='Путь до функции')),
                ('verbose_name', models.CharField(default='Имя функции не определено', max_length=512, verbose_name='Название функции')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Дата и время создания')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Дата и время обновления')),
            ],
            options={
                'verbose_name': 'Зарегистрированная функция',
                'db_table': 'function_tools_registered_function',
            },
        ),
    ]
